import colorlog

__version__ = '1.0.1'

handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter(
    '%(yellow)s%(asctime)s%(reset)s - %(log_color)s%(levelname)s%(reset)s'
    ' - %(message)s',
    '%H:%M:%S'
))
logger = colorlog.getLogger(__name__)
logger.setLevel('INFO')
logger.addHandler(handler)



from datetime import date
from typing import Dict, List, Union
import sys
import uuid

import click
import colorlog
import digitalocean
import toml

from goutte import __version__

log = colorlog.getLogger(__name__)
token = None
error = 0


@click.command(help='DigitalOcean snapshots automation.')
@click.argument('config', envvar='GOUTTE_CONFIG', type=click.File('r'))
@click.argument('do_token', envvar='GOUTTE_DO_TOKEN')
@click.option('--only', type=click.Choice(['snapshot', 'prune']),
              help='Only snapshot or only prune')
@click.option('--debug', is_flag=True, help='Enable debug logging')
@click.version_option(version=__version__)
def entrypoint(config: click.File, do_token: str, only: str,
               debug: bool) -> None:
    """Command line interface entrypoint"""
    global token
    if debug:
        log.setLevel('DEBUG')
    log.info('Starting goutte v{}'.format(__version__))
    token = do_token
    conf = _load_config(config)
    log.debug(f'Retention is set to {conf["retention"]} snapshots')
    if only:
        log.debug(f'Will only {only}')
    _process_droplets(conf, only)
    _process_volumes(conf, only)
    sys.exit(error)


def _load_config(config: click.File) -> Dict[str, Dict]:
    """Return a config dict from a toml config file"""
    try:
        # TODO check minimum validity (retention)
        log.debug('Loading config from {}'.format(config.name))
        conf = toml.load(config)
        assert conf['retention']
        return conf
    except TypeError as e:
        log.critical('Could not read conf {}: {}'.format(config.name, e))
        sys.exit(1)
    except toml.TomlDecodeError as e:
        log.critical('Could not parse toml in config from {}: {}'
                     .format(config.name, e))
        sys.exit(1)
    except KeyError as e:
        log.critical('Malformated configuration: {} is missing'.format(e))
        sys.exit(1)


def _process_droplets(conf: Dict[str, Union[Dict[str, str], str]],
                      only: str) -> None:
    """Execute snapshot and pruning on the droplets"""
    try:
        droplets = _get_droplets(conf['droplets']['names'])
        if droplets:
            log.debug(f'Found {len(droplets)} matching droplets')
            for droplet in droplets:
                log.debug(f'Processing {droplet.name}')
                if only == 'prune' or not only:
                    _prune_droplet_snapshots(droplet, conf['retention'])
                if only == 'snapshot' or not only:
                    _snapshot_droplet(droplet)
        else:
            log.warning('No matching droplet found')
    except KeyError:
        droplets = None
    except KeyboardInterrupt:
        log.critical('Received interuption signal')
        sys.exit(1)


def _process_volumes(conf: Dict[str, Union[Dict[str, str], str]],
                     only: str) -> None:
    """Execute snapshot and pruning on the volumes"""
    try:
        volumes = _get_volumes(conf['volumes']['names'])
        if volumes:
            log.debug(f'Found {len(volumes)} matching volumes')
            for volume in volumes:
                log.debug(f'Processing {volume.name}')
                if only == 'prune' or not only:
                    _prune_volume_snapshots(volume, conf['retention'])
                if only == 'snapshot' or not only:
                    _snapshot_volume(volume)
        else:
            log.warning('No matching volume found')
    except KeyError:
        pass
    except KeyboardInterrupt:
        log.critical('Received interuption signal')
        sys.exit(1)


def _get_droplets(names: List[str]) -> List[digitalocean.Droplet]:
    """Get the droplets objects from the configuration doplets names"""
    try:
        manager = digitalocean.Manager(token=token)
        droplets = manager.get_all_droplets()
        return [droplet for droplet in droplets if droplet.name in names]
    except digitalocean.baseapi.TokenError as e:
        log.error(f'Token not valid: {e}')
    except digitalocean.baseapi.DataReadError as e:
        log.error(f'Could not read response: {e}')
    except digitalocean.baseapi.JSONReadError as e:
        log.error(f'Could not parse json: {e}')
    except digitalocean.baseapi.NotFoundError as e:
        log.error(f'Ressource not found: {e}')
    except Exception as e:
        log.error(f'Unexpected exception: {e}')


def _snapshot_droplet(droplet: digitalocean.Droplet) -> None:
    """Take a snapshot of a given droplet"""
    global error
    name = 'goutte-{}-{}-{}'.format(
        droplet.name,
        date.today().strftime('%Y%m%d'),
        uuid.uuid4().hex[:5])
    try:
        droplet.take_snapshot(name)
        log.info(f'{droplet.name} - Snapshot ({name})')
    except digitalocean.baseapi.TokenError as e:
        log.error(f'Token not valid: {e}.')
        error = 1
    except digitalocean.baseapi.DataReadError as e:
        log.error(f'Could not read response: {e}.')
        error = 1
    except digitalocean.baseapi.JSONReadError as e:
        log.error(f'Could not parse json: {e}.')
        error = 1
    except digitalocean.baseapi.NotFoundError as e:
        log.error(f'Ressource not found: {e}.')
        error = 1
    except Exception as e:
        log.error(f'Unexpected exception: {e}.')
        error = 1


def _prune_droplet_snapshots(droplet: digitalocean.Droplet,
                             retention: int) -> None:
    """Prune goutte snapshots if tmore than the configured retention time"""
    global error
    try:
        all_snapshots = _order_snapshots([
            digitalocean.Snapshot.get_object(
                api_token=token, snapshot_id=snapshot_id
             ) for snapshot_id in droplet.snapshot_ids
        ])
        snapshots = [snapshot for snapshot in all_snapshots
                     if snapshot.name[:6] == 'goutte']
        if len(snapshots) > retention:
            log.debug(f'{droplet.name} - Exceed retention policy by '
                      f'{len(snapshots) - retention}')
            for snapshot in snapshots[:len(snapshots)-retention]:
                log.info(f'{droplet.name} - Prune ({snapshot.name})')
                snapshot.destroy()
    except digitalocean.baseapi.TokenError as e:
        log.error(f'Token not valid: {e}.')
        error = 1
    except digitalocean.baseapi.DataReadError as e:
        log.error(f'Could not read response: {e}.')
        error = 1
    except digitalocean.baseapi.JSONReadError as e:
        log.error(f'Could not parse json: {e}.')
        error = 1
    except digitalocean.baseapi.NotFoundError as e:
        log.error(f'Ressource not found: {e}.')
        error = 1
    except Exception as e:
        log.error(f'Unexpected exception: {e}.')
        error = 1


def _get_volumes(names: List[str]) -> List[digitalocean.Volume]:
    """Get the volumes objects from the configuration volume names"""
    try:
        manager = digitalocean.Manager(token=token)
        volumes = manager.get_all_volumes()
        return [volume for volume in volumes if volume.name in names]
    except digitalocean.baseapi.TokenError as e:
        log.error(f'Token not valid: {e}')
    except digitalocean.baseapi.DataReadError as e:
        log.error(f'Could not read response: {e}')
    except digitalocean.baseapi.JSONReadError as e:
        log.error(f'Could not parse json: {e}')
    except digitalocean.baseapi.NotFoundError as e:
        log.error(f'Ressource not found: {e}')
    except Exception as e:
        log.error(f'Unexpected exception: {e}')


def _snapshot_volume(volume: digitalocean.Volume) -> None:
    """Take a snapshot of a given volume"""
    global error
    name = 'goutte-{}-{}-{}'.format(
        volume.name,
        date.today().strftime('%Y%m%d'),
        uuid.uuid4().hex[:5])
    try:
        volume.snapshot(name)
        log.info(f'{volume.name} - Snapshot ({name})')
    except digitalocean.baseapi.TokenError as e:
        log.error(f'Token not valid: {e}.')
        error = 1
    except digitalocean.baseapi.DataReadError as e:
        log.error(f'Could not read response: {e}.')
        error = 1
    except digitalocean.baseapi.JSONReadError as e:
        log.error(f'Could not parse json: {e}.')
        error = 1
    except digitalocean.baseapi.NotFoundError as e:
        log.error(f'Ressource not found: {e}.')
        error = 1
    except Exception as e:
        log.error(f'Unexpected exception: {e}.')
        error = 1


def _prune_volume_snapshots(volume: digitalocean.Volume,
                            retention: int) -> None:
    """Prune goutte snapshots if tmore than the configured retention time"""
    global error
    try:
        all_snapshots = _order_snapshots(volume.get_snapshots())
        snapshots = [snapshot for snapshot in all_snapshots
                     if snapshot.name[:6] == 'goutte']
        if len(snapshots) > retention:
            log.debug(f'{volume.name} - Exceed retention policy by '
                      f'{len(snapshots) - retention}')
            for snapshot in snapshots[:len(snapshots)-retention]:
                log.info(f'{volume.name} - Prune ({snapshot.name})')
                snapshot.destroy()
    except digitalocean.baseapi.TokenError as e:
        log.error(f'Token not valid: {e}.')
        error = 1
    except digitalocean.baseapi.DataReadError as e:
        log.error(f'Could not read response: {e}.')
        error = 1
    except digitalocean.baseapi.JSONReadError as e:
        log.error(f'Could not parse json: {e}.')
        error = 1
    except digitalocean.baseapi.NotFoundError as e:
        log.error(f'Ressource not found: {e}.')
        error = 1
    except Exception as e:
        log.error(f'Unexpected exception: {e}.')
        error = 1


def _order_snapshots(snapshots: List[digitalocean.Snapshot]
                     ) -> List[digitalocean.Snapshot]:
    """Order snapshots by creation date"""
    return sorted(snapshots, key=lambda x: x.created_at)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['goutte']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'colorlog>=3.1,<4.0',
 'python-digitalocean>=1.14,<2.0',
 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['goutte = goutte.main:entrypoint']}

setup_kwargs = {
    'name': 'goutte',
    'version': '1.0.1',
    'description': 'DigitalOcean snapshot automation service',
    'long_description': '# goutte <a href="https://travis-ci.org/tomochain/goutte"><img align="right" src="https://travis-ci.org/tomochain/goutte.svg?branch=develop"></a>\nDigitalOcean doesn\'t propose any way of automating snapshots.\nThere are [some SaaS](https://snapshooter.io/) that can take care of it but paying to execute some API requests seemed a bit off.\n\nThat\'s why we developed a simple script which can run with cron jobs or in CI services like Travis for free.\nWe use it daily to manage [our backups](https://github.com/tomochain/backups).\n\nIt includes:\n- Snapshoting droplets\n- Snapshoting volumes\n- Retention policy\n- Pruning snapshots\n\n## Requirements\n- Python ^3.6\n- A DigitalOcean account\n\n## Installation\nInstall it directly from pip:\n```bash\npip3 install --user goutte\n```\n\n## Configuration file\nGoutte takes its configuration from a pretty straightforward toml file.\nWe provided and example in `goutte.example.toml`.\n\n```toml\nretention = 10     # Number of backups to keep per droplet/volume\n\n[droplets]\nnames = [          # Array of droplets you want to snapshot\n  \'server01\',\n  \'server02\',\n  \'server03\',\n]\n\n[volumes]\nnames = [          # Array of volumes you want to snapshot\n  \'db01\',\n  \'redis01\',\n  \'redis02\',\n]\n```\n\n## Usage\nGoutte takes two arguments which can also be set via environment variables:\n\n| # | Help     | Description                         | Environment variable |\n| - | -------- | ----------------------------------- | -------------------- |\n| 1 | CONFIG   | Path to the toml configuration file | `GOUTTE_CONFIG`      |\n| 2 | DO_TOKEN | Your DigitalOcean API token         | `GOUTTE_DO_TOKEN`    |\n\n```bash\nUsage: goutte [OPTIONS] CONFIG DO_TOKEN\n\n  DigitalOcean snapshots automation.\n\nOptions:\n  --only [snapshot|prune]  Only snapshot or only prune\n  --debug                  Enable debug logging\n  --version                Show the version and exit.\n  --help                   Show this message and exit.\n```\n\nRunning "snapshot only" for a configuration file containing one droplet and one volume:\n```bash\n$ goutte goutte.toml $do_token --only snapshot\n13:32:48 - INFO - Starting goutte v1.0.1\n13:32:52 - INFO - sgp1-website-01 - Snapshot (goutte-sgp1-website-01-20181220-56bde)\n13:32:59 - INFO - sgp1-mariadb-01 - Snapshot (goutte-sgp1-mariadb-01-20181220-3673d)\n```\n\n## Run with Docker\nWe have a Docker image ready for you to use on Docker Hub.\nIt will read by default the configuration under `/goutte/goutte.toml`\n\n```bash\ndocker run \\\n  -e GOUTTE_DO_TOKEN=${do_token} \\\n  -v $(pwd)/goutte.toml:/goutte/goutte.toml \\\n  tomochain:goutte\n```\n\n## Automating with Travis\nYou can easily automate it via cron job but the easiest way would be by leveraging free CI tools like Travis.\n\n1. You can create a repo which contains your `goutte.toml` configuration and the following travis file `.travis.yml` :\n\n```yml\nlanguage: python\npython: 3.6\n\ninstall:\n  - pip install goutte\n\nscript:\n  - goutte goutte.toml # Don\'t forget to set GOUTTE_DO_TOKEN in Travis config\n```\n\n2. Enable the repo in Travis and then go to the configuration\n3. Add the environment variable GOUTTE_DO_TOKEN with the value of your DigitalOcean API key\n4. Enable daily cron job\n5. You\'re good to go, goutte will run everyday and take care of the snapshots.\n\n**Note**: You can have different retentions for different volumes by having multiple configurations.\n```yml\n# ...\nscript:\n  - goutte 10days.toml\n  - goutte 1day.toml\n```\n\nYou can see how we set it up for ourself [here](https://github.com/tomochain/backups).\n',
    'author': 'etienne-napoleone',
    'author_email': 'etienne@tomochain.com',
    'url': 'https://github.com/tomochain/goutte',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)
